/*
 * Copyright (c) 2004-2008 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2009 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2009 ZIH, TU Dresden, Federal Republic of Germany. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of OpenSM single-source-shortest-path routing
 *    (with dijkstra algorithm)
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif            /* HAVE_CONFIG_H */

#include <stdio.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_node.h>


/* "infinity" for dijkstra */
#define INF      0x7FFFFFFF


enum {
   UNDISCOVERED = 0,
   DISCOVERED
};


typedef struct link {
   uint64_t guid;       /* guid of the neighbor behind the link */
   uint32_t from;       /* base_index in the adjazenz list (start of the link) */
   uint8_t  from_port;  /* port on the base_side (needed for weight update to identify the correct link for multigraphs) */
   uint32_t to;         /* index of the neighbor in the adjazenz list (end of the link) */
   uint8_t  to_port;    /* port on the side of the neighbor (needed for the LFT) */
   uint64_t weight;     /* link weight */
   struct link *next;
} link_t;

typedef struct vertex {
   /* informations of the fabric */
   uint64_t  guid;
   uint16_t  lid;       /* for lft filling */
   uint32_t  num_hca;   /* numbers of Hca on the switch, for weight calculation */
   link_t   *links;
   uint8_t   hops;
   /* for dijkstra routing */
   link_t   *used_link; /* link between the vertex discovered before and this vertex */
   uint64_t  distance;  /* distance from source to this vertex*/
   uint8_t   state;
   /* for the binary heap */
   uint32_t  heap_id;
   /* for LFT writing and debug */
   osm_switch_t *sw;    /* selfpointer */
} vertex_t;


typedef struct binary_heap {
   uint32_t size;       /* size of the heap */
   vertex_t **nodes;    /* array with pointers to elements of the adj_list */
} binary_heap_t;


typedef struct sssp_context {
   osm_ucast_mgr_t *p_mgr;
   vertex_t        *adj_list;
   uint32_t         adj_list_size;
} sssp_context_t;


/**************** set initial values for structs **********************
 **********************************************************************/
static inline void set_default_link(link_t *link)
{
   link->guid = 0;
   link->from = 0;
   link->from_port = 0;
   link->to = 0;
   link->to_port = 0;
   link->weight = 0;
   link->next = NULL;
}


static inline void set_default_vertex(vertex_t *vertex)
{
   vertex->guid = 0;
   vertex->lid = 0;
   vertex->num_hca = 0;
   vertex->links = NULL;
   vertex->hops = 0;
   vertex->used_link = NULL;
   vertex->distance = 0;
   vertex->state = UNDISCOVERED;
   vertex->heap_id = 0;
   vertex->sw = NULL;
}
/**********************************************************************
 **********************************************************************/


/************ helper functions for heap in dijkstra *******************
 **********************************************************************/
/* returns true if element 1 is smaller than element 2 */
static inline uint32_t heap_smaller(binary_heap_t *heap, uint32_t i, uint32_t j)
{
   return (heap->nodes[i]->distance < heap->nodes[j]->distance) ? 1 : 0;
}


/* swap two elements */
static void heap_exchange(binary_heap_t *heap, uint32_t i, uint32_t j)
{
   uint32_t tmp_heap_id = 0;
   vertex_t *tmp_node = NULL;

   /* 1. swap the heap_id */
   tmp_heap_id = heap->nodes[i]->heap_id;
   heap->nodes[i]->heap_id = heap->nodes[j]->heap_id;
   heap->nodes[j]->heap_id = tmp_heap_id;
   /* 2. swap pointers */
   tmp_node = heap->nodes[i];
   heap->nodes[i] = heap->nodes[j];
   heap->nodes[j] = tmp_node;
}


/* changes position of element with parent until children are bigger */
static uint32_t heap_up(binary_heap_t *heap, uint32_t i)
{
   uint32_t curr = i, father = 0;
   
   if (curr > 0) {
      father = (curr - 1) >> 1;
      while (heap_smaller(heap, curr, father)) {
         heap_exchange(heap, curr, father);
         /* try to go up when we arent already root */
         curr = father;
         if (curr > 0)
            father = (curr - 1) >> 1;
      }
   }
   
   return curr;
}


/* changes position of element with children until parent is smaller */
static uint32_t heap_down(binary_heap_t *heap, uint32_t i)
{
   uint32_t curr = i;
   uint32_t son1 = 0, son2 = 0, smaller_son = 0;
   uint32_t exchanged = 0;
   
   do {
      son1 = ((curr + 1) << 1) - 1;
      son2 = (curr + 1) << 1;
      exchanged = 0;
      
      /* exchange with smaller son */
      if (son1 < heap->size && son2 < heap->size) {
         if (heap_smaller(heap, son1, son2))
            smaller_son = son1;
         else
            smaller_son = son2;
      } else if (son1 < heap->size) {
         /* only one son */
         smaller_son = son1;
      } else {
         /* finished */
         break;
      }
      
      /* only exchange when smaller */
      if (heap_smaller(heap, smaller_son, curr)) {
         heap_exchange(heap, curr, smaller_son);
         exchanged = 1;
         curr = smaller_son;
      }
   } while (exchanged);
   
   return curr;
}


/* reheapify element */
static inline void heap_heapify(binary_heap_t *heap, uint32_t i)
{
   heap_down(heap, heap_up(heap, i));
}


/* creates heap for graph */
static int heap_create(vertex_t *adj_list, uint32_t adj_list_size, binary_heap_t **binheap)
{
   binary_heap_t *heap = NULL;
   uint32_t i = 0;
   
   /* allocate the memory for the heap object */
   heap = (binary_heap_t *)malloc(sizeof(binary_heap_t));
   if (!heap)
      return 1;
   
   /* the heap size is equivalent to the size of the adj_list */
   heap->size = adj_list_size;
   
   /* allocate the pointer array, fill with the pointers to the elements of the adj_list and set the initial heap_id */
   heap->nodes = (vertex_t **)malloc(heap->size * sizeof(vertex_t *));
   if (!heap->nodes) {
      free(heap);
      return 1;
   }
   for (i = 0; i < heap->size; i++) {
      heap->nodes[i] = &adj_list[i];
      heap->nodes[i]->heap_id = i;
   }
   
   /* sort elements */
   for (i = heap->size; i > 0; i--)
      heap_down(heap, i - 1);
   
   *binheap = heap;
   return 0;
}


/* returns current minimum and removes it from heap */
static vertex_t * heap_getmin(binary_heap_t *heap)
{
   vertex_t *min = NULL;
   
   if (heap->size > 0)
      min = heap->nodes[0];
   
   if (min == NULL)
      return min;
   
   if (heap->size > 0) {
      if (heap->size > 1) {
         heap_exchange(heap, 0, heap->size - 1);
         heap->size--;
         heap_down(heap, 0);
      } else {
         heap->size--;
      }
   }
   
   return min;
}


/* cleanup heap */
static void heap_free(binary_heap_t *heap)
{
   if (heap) {
      if (heap->nodes) {
         free(heap->nodes);
         heap->nodes = NULL;
      }
      free(heap);
   }
}
/**********************************************************************
 **********************************************************************/


static void sssp_print_graph(osm_ucast_mgr_t *p_mgr, vertex_t *adj_list, uint32_t size)
{
   uint32_t i = 0, c = 0;
   link_t *link = NULL;

   /* index 0 is for the source in dijkstra -> ignore */
   for (i = 1; i < size; i++) {
      OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "adj_list[%"PRIu32"]:\n", i);
      OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "   guid = 0x%"PRIx64" lid = %"PRIu16" (%s)\n", adj_list[i].guid, adj_list[i].lid, adj_list[i].sw->p_node->print_desc);
      OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "   num_hca = %"PRIu32"\n", adj_list[i].num_hca);

      c = 1;
      for (link = adj_list[i].links; link != NULL; link = link->next, c++) {
         OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "   link[%"PRIu32"]:\n", c);
         OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "      to guid = 0x%"PRIx64" (%s) port %"PRIu8"\n", link->guid, adj_list[link->to].sw->p_node->print_desc, link->to_port);
         OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "      weight on this link = %"PRIu64"\n", link->weight);
      }
   }
}


/* traverse subnet to gather information about the connected switches */
static int sssp_build_graph(void *context)
{
   sssp_context_t *sssp_ctx = (sssp_context_t *)context;
   osm_ucast_mgr_t *p_mgr = (osm_ucast_mgr_t *)(sssp_ctx->p_mgr);
   
   cl_qmap_t *port_tbl = &p_mgr->p_subn->port_guid_tbl; /* 1 managment port per switch + 1 or 2 ports for each Hca */
   osm_port_t *p_port = NULL;
   cl_qmap_t *sw_tbl = &p_mgr->p_subn->sw_guid_tbl;
   cl_map_item_t *item = NULL;
   osm_switch_t *sw = NULL;
   osm_node_t *remote_node = NULL;
   uint8_t port = 0, remote_port = 0;
   uint32_t i = 0, j = 0;
   uint32_t total_num_hca = 0;
   vertex_t *adj_list = NULL;
   osm_physp_t *p_physp = NULL;
   link_t *link = NULL, *head = NULL;
   uint32_t num_sw = 0, adj_list_size = 0;
   
   OSM_LOG_ENTER(p_mgr->p_log);
   OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE, "Building graph for sssp routing\n");
   
   num_sw = cl_qmap_count(sw_tbl);
   adj_list_size = num_sw + 1;
   /* allocate an adjazenz list (array), 0. element is reserved for the source (Hca) in the routing algo, others are switches */
   adj_list = (vertex_t *)malloc(adj_list_size * sizeof(vertex_t));
   if (!adj_list) {
      OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR AD02: cannot allocate memory for adj_list\n");
      return 1;
   }
   for (i = 0; i < adj_list_size; i++)
      set_default_vertex(&adj_list[i]);
   
   /* count the total number of Hca in the fabric */
   for (item = cl_qmap_head(port_tbl); item != cl_qmap_end(port_tbl); item = cl_qmap_next(item)) {
      p_port = (osm_port_t *)item;
      if (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_CA)
         total_num_hca++;
   }

   i = 1;   /* fill adj_list -> start with index 1 */
   for (item = cl_qmap_head(sw_tbl); item != cl_qmap_end(sw_tbl); item = cl_qmap_next(item), i++) {
      sw = (osm_switch_t *)item;
      OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Processing switch with GUID 0x%"PRIx64"\n", cl_ntoh64(osm_node_get_node_guid(sw->p_node)));
      
      adj_list[i].guid = cl_ntoh64(osm_node_get_node_guid(sw->p_node));
      adj_list[i].lid = cl_ntoh16(osm_node_get_base_lid(sw->p_node, 0));
      adj_list[i].sw = sw;
      
      link = (link_t *)malloc(sizeof(link_t));
      if (!link) {
         OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR AD03: cannot allocate memory for a link\n");
         return 1;
      }
      head = link;
      head->next = NULL;
      
      /* iterate over all ports in the switch, start with port 1 (port 0 is a managment port) */
      for (port = 1; port < sw->num_ports; port++) {
         /* get the node behind the port */
         remote_node = osm_node_get_remote_node(sw->p_node, port, &remote_port);
         /* if there is no remote node on this port or it's the same switch -> try next port */
         if (!remote_node || remote_node->sw == sw)
            continue;
         /* make sure the link is healthy */
         p_physp = osm_node_get_physp_ptr(sw->p_node, port);
         if (!p_physp || !osm_link_is_healthy(p_physp))
            continue;
         /* if there is a Hca connected -> count and cycle */
         if (!remote_node->sw) {
            adj_list[i].num_hca++;
            continue;
         }
         OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Node 0x%"PRIx64", remote node 0x%"PRIx64", port %"PRIu8", remote port %"PRIu8"\n", cl_ntoh64(osm_node_get_node_guid(sw->p_node)), cl_ntoh64(osm_node_get_node_guid(remote_node)), port, remote_port);
         
         link->next = (link_t *)malloc(sizeof(link_t));
         if (!link->next) {
            OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR AD08: cannot allocate memory for a link\n");
            return 1;
         }
         link = link->next;
         set_default_link(link);
         link->guid = cl_ntoh64(osm_node_get_node_guid(remote_node));
         link->from = i;
         link->from_port = port;
         link->to_port = remote_port;
         link->weight = total_num_hca * total_num_hca; /* initilize with P^2 to force shortest paths */
      }
      
      adj_list[i].links = head->next;
      free(head);
   }
   /* connect the links with it's second adjacent node in the list */
   for (i = 1; i < adj_list_size; i++) {
      link = adj_list[i].links;
      while (link) {
         for (j = 1; j < adj_list_size; j++) {
            if (link->guid == adj_list[j].guid) {
               link->to = j;
               break;
            }
         }
         link = link->next;
      }
   }
   /* print the discovered graph */
   if (osm_log_get_level(p_mgr->p_log) >= OSM_LOG_DEBUG)
      sssp_print_graph(p_mgr, adj_list, adj_list_size);
   
   sssp_ctx->adj_list = adj_list;
   sssp_ctx->adj_list_size = adj_list_size;
   
   OSM_LOG_EXIT(p_mgr->p_log);
   return 0;
}


static void print_routes(osm_ucast_mgr_t *p_mgr, vertex_t *adj_list, uint32_t adj_list_size, osm_port_t *port)
{
   uint32_t i = 0, j = 0;

   for (i = 1; i < adj_list_size; i++) {
      if (adj_list[i].state == DISCOVERED) {
         OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Route from 0x%"PRIx64" (%s) to 0x%"PRIx64" (%s):\n", adj_list[i].guid, adj_list[i].sw->p_node->print_desc, cl_ntoh64(osm_node_get_node_guid(port->p_node)), port->p_node->print_desc);
         j = i;
         while (adj_list[j].used_link) {
            if (j > 0){
               OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "   0x%"PRIx64" (%s) routes thru port %"PRIu8"\n", adj_list[j].guid, adj_list[j].sw->p_node->print_desc, adj_list[j].used_link->to_port);
            } else {
               OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "   0x%"PRIx64" (%s) routes thru port %"PRIu8"\n", adj_list[j].guid, port->p_node->print_desc, adj_list[j].used_link->to_port);
            }
            j = adj_list[j].used_link->from;
         }
      }
   }
}


/* dijkstra step from one source to all switches in the sssp graph */
static int dijkstra(osm_ucast_mgr_t *p_mgr, vertex_t *adj_list, uint32_t adj_list_size, osm_port_t *port)
{
   uint32_t i = 0, j = 0, index = 0;
   osm_node_t *remote_node = NULL;
   uint8_t remote_port = 0;
   vertex_t *current = NULL;
   link_t *link = NULL;
   uint64_t guid = 0;
   binary_heap_t *heap = NULL;
   int err = 0;
   
   OSM_LOG_ENTER(p_mgr->p_log);
   
   /* reset all switches for new round with a new source for dijkstra */
   for (i = 1; i < adj_list_size; i++) {
      adj_list[i].hops = 0;
      adj_list[i].used_link = NULL;
      adj_list[i].distance = INF;
      adj_list[i].state = UNDISCOVERED;
   }
   
   /* if behind port is a Hca -> set adj_list[0] */
   if (osm_node_get_type(port->p_node) == IB_NODE_TYPE_CA) {
      /* save old link to prevent many mallocs after set_default_... */
      link = adj_list[0].links;
      /* initialize adj_list[0] (the source for the routing, a Hca) */
      set_default_vertex(&adj_list[0]);
      adj_list[0].guid = cl_ntoh64(osm_node_get_node_guid(port->p_node));
      adj_list[0].lid = cl_ntoh16(osm_port_get_base_lid(port));
      index = 0;
      /* write saved link back to new adj_list[0] */
      adj_list[0].links = link;
      
      /* initialize link to neighbor for adj_list[0];
         make sure the link is healthy
      */
      if (port->p_physp && osm_link_is_healthy(port->p_physp)) {
         remote_node = osm_node_get_remote_node(port->p_node, port->p_physp->port_num, &remote_port);
         /* if there is no remote node on this port or it's the same Hca -> ignore */
         if (remote_node && (osm_node_get_type(remote_node) == IB_NODE_TYPE_SWITCH)) {
            if ( !(adj_list[0].links) ) {
               adj_list[0].links = (link_t *)malloc(sizeof(link_t));
               if (!(adj_list[0].links)) {
                  OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR AD07: cannot allocate memory for a link\n");
                  return 1;
               }
            }
            set_default_link(adj_list[0].links);
            adj_list[0].links->guid = cl_ntoh64(osm_node_get_node_guid(remote_node));
            adj_list[0].links->from_port = port->p_physp->port_num;
            adj_list[0].links->to_port = remote_port;
            adj_list[0].links->weight = 1;
            for (j = 1; j < adj_list_size; j++) {
               if (adj_list[0].links->guid == adj_list[j].guid) {
                  adj_list[0].links->to = j;
                  break;
               }
            }
         }
      }
   /* if behind port is a switch -> search switch in adj_list */
   } else {
      /* reset adj_list[0], if links=NULL reset was done before, then skip */
      if (adj_list[0].links) {
         free(adj_list[0].links);
         set_default_vertex(&adj_list[0]);
      }
      /* search for the switch which is the source in this round */
      guid = cl_ntoh64(osm_node_get_node_guid(port->p_node));
      for (i = 1; i < adj_list_size; i++) {
         if (guid == adj_list[i].guid) {
            index = i;
            break;
         }
      }
   }
   
   /* source in dijkstra */
   adj_list[index].distance = 0;
   adj_list[index].state = DISCOVERED;
   adj_list[index].hops = 0; /* the source has hop count = 0 */
   
   /* create a heap to find (efficient) the node with the smallest distance */
   if (osm_node_get_type(port->p_node) == IB_NODE_TYPE_CA)
      err = heap_create(adj_list, adj_list_size, &heap);
   else
      err = heap_create(&adj_list[1], adj_list_size - 1, &heap);
   if (err) {
      OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR AD09: cannot allocate memory for heap or heap->node in heap_create(...)\n");
      return err;
   }
   
   current = heap_getmin(heap);
   while (current) {
      current->state = DISCOVERED;
      if (current->used_link) /* increment the number of hops to the source for each new node */
         current->hops = adj_list[current->used_link->from].hops + 1;
      
      /* add/update nodes which aren't discovered but accessible */
      for (link = current->links; link != NULL; link = link->next) {
         if ((adj_list[link->to].state != DISCOVERED) && (current->distance + link->weight < adj_list[link->to].distance)) {
            adj_list[link->to].used_link = link;
            adj_list[link->to].distance = current->distance + link->weight;
            heap_heapify(heap, adj_list[link->to].heap_id);
         }
      }
      
      current = heap_getmin(heap);
   }
   
   /* destroy the heap */
   heap_free(heap);
   heap = NULL;
   
   OSM_LOG_EXIT(p_mgr->p_log);
   return 0;
}


/* update the linear forwarding tables of all switches with the informations
   from the last dijsktra step
*/
static int update_lft(osm_ucast_mgr_t *p_mgr, vertex_t *adj_list, uint32_t adj_list_size, osm_port_t *p_port)
{
   uint32_t i = 0;
   int32_t index = -1;
   uint64_t guid;
   uint16_t lid = 0;
   uint8_t port = 0;
   uint8_t hops = 0;
   osm_switch_t *p_sw = NULL;
   boolean_t is_ignored_by_port_prof = FALSE;
   osm_physp_t *p = NULL;
   cl_status_t ret;
   
   OSM_LOG_ENTER(p_mgr->p_log);
   
   if (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_SWITCH) {
      /* we have to search for the right switch, with it's lid to update the LFT */
      guid = cl_ntoh64(osm_node_get_node_guid(p_port->p_node));
      for (i = 0; i < adj_list_size; i++) {
         if (adj_list[i].guid == guid) {
            index = i;
            break;
         }
      }
   } else {
      /* update the routing to a Hca -> index 0 contains the Hca */
      index = 0;
   }
   
   if (index >= 0) {
      lid = adj_list[index].lid;
   } else { 
      OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR AD06: cannot find port in adj_list to run update_lft\n");
      return 1;
   }
   
   for (i = 1; i < adj_list_size; i++) {
      /* for each switch the port to the 'self'lid is the managment port (=0) */
      adj_list[i].sw->new_lft[adj_list[i].lid] = 0;
      /* the hop count to to the 'self'lid is 0 for each switch */
      osm_switch_set_hops(adj_list[i].sw, adj_list[i].lid, 0, 0);
      
      /* if no route goes thru this switch -> cycle */
      if ( !(adj_list[i].used_link) )
         continue;
      
      p_sw = adj_list[i].sw;
      hops = adj_list[i].hops;
      port = adj_list[i].used_link->to_port;
      /* the used_link is the link that was used in dijkstra to reach this node,
         so the to_port is the local port on this node
      */
      
      if (port == OSM_NO_PATH) { /* if clause shouldn't be possible in this routing, but who cares */
         OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "No path to get to LID %"PRIu16" from switch 0x%"PRIx64"\n", lid, cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
         
         /* do not try to overwrite the ppro of non existing port ... */
         is_ignored_by_port_prof = TRUE;
      } else {
         OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Routing LID %"PRIu16" to port %"PRIu8" for switch 0x%"PRIx64"\n", lid, port, cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
         
         p = osm_node_get_physp_ptr(p_sw->p_node, port);
         
         /* we would like to optionally ignore this port in equalization
            as in the case of the Mellanox Anafa Internal PCI TCA port
          */
         is_ignored_by_port_prof = p->is_prof_ignored;
         
         /* We also would ignore this route if the target lid is of
            a switch and the port_profile_switch_node is not TRUE
          */
         if (!p_mgr->p_subn->opt.port_profile_switch_nodes)
            is_ignored_by_port_prof |= (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_SWITCH);
      }
      
      /* set port in LFT */
      p_sw->new_lft[lid] = port;
      if (!is_ignored_by_port_prof) {
         /* update the number of path routing thru this port */
         osm_switch_count_path(p_sw, port);
      }
      /* set te hop count from this switch to the lid */
      ret = osm_switch_set_hops(p_sw, lid, port, hops);
      if (ret != CL_SUCCESS)
         OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR AD05: cannot set hops for LID %"PRIu16" at switch 0x%"PRIx64"\n", lid, cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
   }
   
   OSM_LOG_EXIT(p_mgr->p_log);
   return 0;
}


/* increment the edge weights of the sssp graph which represent the number
   of paths on this link
*/
static void update_weights(osm_ucast_mgr_t *p_mgr, vertex_t *adj_list, uint32_t adj_list_size, osm_port_t *port)
{
   uint32_t i = 0, j = 0;
   uint32_t additional_weight = 0;
   
   OSM_LOG_ENTER(p_mgr->p_log);
   
   for (i = 1; i < adj_list_size; i++) {
      /* if no route goes thru this switch -> cycle */
      if ( !(adj_list[i].used_link) )
         continue;
      /* if the source of dijkstra was a Hca -> add num_hca for the weight, else a weight of 1 */
      if (osm_node_get_type(port->p_node) == IB_NODE_TYPE_CA)
         additional_weight = adj_list[i].num_hca;
      else
         additional_weight = 1;
      
      j = i;
      while (adj_list[j].used_link) {
         /* update the link from pre to this node */
         adj_list[j].used_link->weight += additional_weight;
         
         j = adj_list[j].used_link->from;
      }
   }
   
   OSM_LOG_EXIT(p_mgr->p_log);
}


/* meta function which calls subfunctions for dijkstra, update lft and weights
   to calculate the routing for the subnet
*/
static int sssp_do_dijkstra_routing(void *context)
{
   sssp_context_t *sssp_ctx = (sssp_context_t *)context;
   osm_ucast_mgr_t *p_mgr = (osm_ucast_mgr_t *)sssp_ctx->p_mgr;
   vertex_t *adj_list = (vertex_t *)sssp_ctx->adj_list;
   uint32_t adj_list_size = sssp_ctx->adj_list_size;
   
   cl_qmap_t *port_tbl = &p_mgr->p_subn->port_guid_tbl; /* 1 managment port per switch + 1 or 2 ports for each Hca */
   cl_qmap_t *sw_tbl = &p_mgr->p_subn->sw_guid_tbl;
   cl_map_item_t *item = NULL;
   osm_switch_t *sw = NULL;
   osm_port_t *port = NULL;
   uint32_t i = 0, err = 0;
   
   OSM_LOG_ENTER(p_mgr->p_log);
   OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE, "Calculating shortest path from all Hca/switches to all\n");
   
   /* reset the new_lft for each switch */
   for (item = cl_qmap_head(sw_tbl); item != cl_qmap_end(sw_tbl); item = cl_qmap_next(item)) {
      sw = (osm_switch_t*)item;
      /* initialize LIDs in buffer to invalid port number */
      memset(sw->new_lft, OSM_NO_PATH, sw->max_lid_ho + 1);
   }
   
   /* do the routing for the each Hca in the subnet */
   for (item = cl_qmap_head(port_tbl); item != cl_qmap_end(port_tbl); item = cl_qmap_next(item)) {
      port = (osm_port_t *)item;
      
      /* if behind port is a Hca -> calculate shortest path with dijkstra from node to all switches/Hca */
      if (osm_node_get_type(port->p_node) == IB_NODE_TYPE_CA) {
         OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Processing Hca with GUID 0x%"PRIx64"\n", cl_ntoh64(osm_node_get_node_guid(port->p_node)));

         /* do dijkstra from this Hca to each switch */
         err = dijkstra(p_mgr, adj_list, adj_list_size, port);
         if (err)
            return err;
         if (osm_log_get_level(p_mgr->p_log) >= OSM_LOG_DEBUG)
            print_routes(p_mgr, adj_list, adj_list_size, port);

         /* make an update for the linear forwarding tables of the switches */
         err = update_lft(p_mgr, adj_list, adj_list_size, port);
         if (err)
            return err;

         /* add weights for calculated routes to adjust the weights for the next cycle */
         update_weights(p_mgr, adj_list, adj_list_size, port);

         if (osm_log_get_level(p_mgr->p_log) >= OSM_LOG_DEBUG)
            sssp_print_graph(p_mgr, adj_list, adj_list_size);
      }
   }
   /* do the routing for the each switch in the subnet to add the routes from switch to switch */
   for (item = cl_qmap_head(port_tbl); item != cl_qmap_end(port_tbl); item = cl_qmap_next(item)) {
      port = (osm_port_t *)item;
      
      if (osm_node_get_type(port->p_node) == IB_NODE_TYPE_SWITCH) {
         OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Processing switch with GUID 0x%"PRIx64"\n", cl_ntoh64(osm_node_get_node_guid(port->p_node)));

         /* do dijkstra from this switch to each switch */
         err = dijkstra(p_mgr, adj_list, adj_list_size, port);
         if (err)
            return err;
         if (osm_log_get_level(p_mgr->p_log) >= OSM_LOG_DEBUG)
            print_routes(p_mgr, adj_list, adj_list_size, port);

         /* make an update for the linear forwarding tables of the switches */
         err = update_lft(p_mgr, adj_list, adj_list_size, port);
         if (err)
            return err;

         /* add weights for calculated routes to adjust the weights for the next cycle */
         update_weights(p_mgr, adj_list, adj_list_size, port);

         if (osm_log_get_level(p_mgr->p_log) >= OSM_LOG_DEBUG)
            sssp_print_graph(p_mgr, adj_list, adj_list_size);
      }
   }
   
   /* print the new_lft for each switch after routing is done */
   if (osm_log_get_level(p_mgr->p_log) >= OSM_LOG_DEBUG) {
      for (item = cl_qmap_head(sw_tbl); item != cl_qmap_end(sw_tbl); item = cl_qmap_next(item)) {
         sw = (osm_switch_t*)item;
         OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Summery of the (new) LFT for switch 0x%"PRIx64" (%s):\n", cl_ntoh64(osm_node_get_node_guid(sw->p_node)), sw->p_node->print_desc);
         for (i = 0; i < sw->max_lid_ho + 1; i++)
            if (sw->new_lft[i] != OSM_NO_PATH) {
               OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "   for LID=%"PRIu32" use port=%"PRIu8"\n", i, sw->new_lft[i]);
            }
      }
   }
   
   OSM_LOG_EXIT(p_mgr->p_log);
   return 0;
}


static sssp_context_t * sssp_context_create(osm_opensm_t *p_osm)
{
   sssp_context_t *sssp_ctx = NULL;
   
   /* allocate memory */
   sssp_ctx = (sssp_context_t *)malloc(sizeof(sssp_context_t));
   if (sssp_ctx) {
      /* set initial values */
      sssp_ctx->p_mgr = (osm_ucast_mgr_t *)&(p_osm->sm.ucast_mgr);
      sssp_ctx->adj_list = NULL;
   } else {
      OSM_LOG(p_osm->sm.ucast_mgr.p_log, OSM_LOG_ERROR, "ERR AD04: cannot allocate memory for sssp_ctx in sssp_context_create\n");
      return NULL;
   }
   
   return sssp_ctx;
}


static void sssp_context_destroy(void *context)
{
   sssp_context_t *sssp_ctx = (sssp_context_t *)context;
   vertex_t *adj_list = (vertex_t *)(sssp_ctx->adj_list);
   uint32_t i = 0;
   link_t *link = NULL, *tmp=NULL;
   
   for (i = 0; i < sssp_ctx->adj_list_size; i++) {
      link = adj_list[i].links;
      while (link) {
         tmp = link;
         link = link->next;
         free(tmp);
      }
   }
   free(adj_list);
   sssp_ctx->adj_list = NULL;
}


static void delete(void *context)
{
   if (!context)
      return;
   sssp_context_destroy(context);
}


int osm_ucast_sssp_setup(struct osm_routing_engine *r, osm_opensm_t *p_osm)
{
   /* create context container and add ucast managment object */
   sssp_context_t *sssp_context = sssp_context_create(p_osm);
   if (!sssp_context) {
      return 1; /* alloc failed -> skip this routing */
   }
   
   /* reset function pointers to sssp routines */
   r->context = (void *)sssp_context;
   r->build_lid_matrices = sssp_build_graph;
   r->ucast_build_fwd_tables = sssp_do_dijkstra_routing;
   r->delete = delete;
   
   return 0;
}

